<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace EventcorrelationModule\Model\Table;

use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\FileDebugger;

/**
 * Eventcorrelations Model
 *
 * @property \App\Model\Table\HostsTable|\Cake\ORM\Association\BelongsTo $Hosts
 * @property \App\Model\Table\ServicesTable|\Cake\ORM\Association\BelongsTo $Services
 *
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation get($primaryKey, $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation newEntity($data = null, array $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation[] newEntities(array $data, array $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation|bool save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation[] patchEntities($entities, array $data, array $options = [])
 * @method \EventcorrelationModule\Model\Entity\Eventcorrelation findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TreeBehavior
 */
class EventcorrelationsTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('eventcorrelations');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Tree', [
            'scope' => 'Eventcorrelations.host_id',
        ]);

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts'
        ]);

        $this->belongsTo('Services', [
            'foreignKey' => 'service_id',
            'joinType'   => 'INNER',
            'className'  => 'Services'
        ]);

    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', 'create');

        $validator
            ->scalar('operator')
            ->maxLength('operator', 12)
            ->allowEmptyString('operator');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn(['host_id'], 'Hosts'));
        $rules->add($rules->existsIn(['service_id'], 'Services'));

        return $rules;
    }


    /**
     * @param int $evcHostId
     * @return array
     */
    public function getEvcHostsById($evcHostId) {
        $query = $this->find();
        $query->contain([
            'Services' => function (Query $query) {
                $query->disableAutoFields()
                    ->select([
                        'Services.id',
                        'Services.host_id'
                    ]);
                return $query;
            }
        ])
            ->where([
                'Eventcorrelations.host_id' => $evcHostId
            ])
            ->all();

        return $query->toArray();
    }

    /**
     * @param $evcHostId
     * @return array|\Cake\Datasource\EntityInterface
     * @throws RecordNotFoundException
     */
    public function getRootelementByHostId($evcHostId) {
        $query = $this->find();
        $query->contain([
            'Hosts' => function (Query $query) {
                $query->disableAutoFields()
                    ->select([
                        'Hosts.id',
                        'Hosts.name',
                        'Hosts.uuid',
                        'Hosts.container_id'
                    ]);
                return $query;
            }
        ])
            ->where([
                'Eventcorrelations.host_id' => $evcHostId
            ]);

        return $query->firstOrFail();
    }

    /**
     * @param int $evcHostId
     * @param array $hostIds
     * @return array
     */
    public function getEvcTreeData($evcHostId, $hostIds) {
        try {
            $rootElement = $this->getRootelementByHostId($evcHostId);
        } catch (RecordNotFoundException $e) {
            //No rootElement found - return empty EVC tree
            return [];
        }

        $servicesWhere = [];
        if (!empty($hostIds)) {
            if (!is_array($hostIds)) {
                $hostIds = [$hostIds];
            }

            $servicesWhere['Hosts.id IN'] = $hostIds;
        }

        $query = $this->find('threaded');
        $query->contain([
            'Hosts'    => function (Query $query) {
                $query
                    ->disableAutoFields()
                    ->select([
                        'Hosts.id',
                        'Hosts.name'
                    ]);
                return $query;
            },
            'Services' => function (Query $query) use ($servicesWhere) {
                $query
                    ->disableAutoFields()
                    ->select([
                        'Services.id',
                        'Services.servicetemplate_id',
                        'Services.host_id',
                        'Services.name',
                        'Services.description',
                        'Services.service_type',
                        'Services.uuid',
                        'Services.disabled',
                        'servicename' => $query->newExpr('IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)'),
                    ])
                    ->where($servicesWhere)
                    ->contain([
                        'Servicetemplates' => function (Query $query) {
                            $query
                                ->disableAutoFields()
                                ->select([
                                    'Servicetemplates.id',
                                    'Servicetemplates.name',
                                    'Servicetemplates.description'
                                ]);
                            return $query;
                        },
                        'Hosts'            => function (Query $query) {
                            $query
                                ->disableAutoFields()
                                ->select([
                                    'Hosts.id',
                                    'Hosts.name',
                                ]);
                            return $query;
                        }
                    ]);
                return $query;
            }
        ])
            ->where([
                'Eventcorrelations.host_id' => $evcHostId
            ])
            ->disableHydration();

        $childElements = $query->toArray();
        $evcTree = $this->reorganizeEvcStructure($childElements);
        $evcTree = array_reverse($evcTree);

        //Replace the "null" / "" empty string array index
        foreach ($evcTree as $index => $data) {
            foreach ($data as $jsonKey => $evcLayer) {
                if ($jsonKey === '') {
                    unset($evcTree[$index]['']);
                    // Only services where parent_id === null have the suffix _vService in the key of the JSON tree
                    // This was the "" empty key service in V3 !! mindblown
                    $evcTree[$index][$evcLayer[0]['id'] . '_vService'] = $evcLayer; //ROOT_NODE // $evcLayer[0]['id']
                }
            }
        }


        return $evcTree;
    }

    /**
     * @param $evcAsNest
     * @return array
     */
    public function reorganizeEvcStructure($evcAsNest) {
        $result = [];
        $this->flattenElement($evcAsNest, 1, $result);
        return $result;
    }

    /**
     * @param $children
     * @param $layer
     * @param $result
     */
    private function flattenElement($children, $layer, &$result) {
        foreach ($children as $element) {
            $data = [
                'id'         => $element['id'],
                'parent_id'  => $element['parent_id'],
                'host_id'    => $element['host_id'],
                'service_id' => $element['service_id'],
                'operator'   => $element['operator'],
                'service'    => $element['service'],
            ];
            if (isset($element['servicename'])) {
                $data['service']['servicename'] = $element['servicename'];
            }

            $result[$layer][$element['parent_id']][] = $data;
        }
        foreach ($children as $element) {
            $this->flattenElement($element['children'], $layer + 1, $result);
        }
    }


    public function getFirstLayerServicesByEvcHostId($evcHostId) {
        $query = $this->find();
        $query
            ->contain([
                'Services' => function (Query $query) {
                    $query
                        ->disableAutoFields()
                        ->select([
                            'Services.id',
                            'Services.host_id',
                            'Services.uuid',
                        ]);
                    return $query;
                }
            ])
            ->where([
                'Eventcorrelations.host_id' => $evcHostId,
                'Eventcorrelations.operator IS NULL'
            ])
            ->all();

        return $query->toArray();
    }

    /**
     * @param array $hostIds
     * @return array
     *
     * Returns all hosts and services that are used in another event correlation by evcHostIds
     */
    public function inUse($hostIds = []) {
        if (!is_array($hostIds)) {
            $hostIds = [$hostIds];
        }

        if (empty($hostIds)) {
            return [];
        }

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $query = $HostsTable->find()
            ->select([
                'Hosts.id',
                'Hosts.name',
                'Hosts.description',
                'Services.id',
                'Services.name',
                'Services.service_type',
                'Services.host_id',
                'Eventcorrelations.service_id',
                'Eventcorrelations.host_id',
                'EvcHosts.id',
                'EvcHosts.name',
                'EvcHosts.description',
                'EvcHosts.container_id'
            ])
            ->join([
                [
                    'table'      => 'services',
                    'alias'      => 'Services',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Services.host_id = Hosts.id',
                        'Services.service_type' => EVK_SERVICE
                    ],
                ],
                [
                    'table'      => 'eventcorrelations',
                    'alias'      => 'Eventcorrelations',
                    'type'       => 'LEFT',
                    'conditions' => [
                        'AND' => [
                            'Eventcorrelations.service_id = Services.id',
                            'Eventcorrelations.host_id != Services.host_id',
                        ]
                    ],
                ],
                [
                    'table'      => 'hosts',
                    'alias'      => 'EvcHosts',
                    'type'       => 'LEFT',
                    'conditions' => [
                        'EvcHosts.id = Eventcorrelations.host_id',
                    ],
                ],
            ])
            ->where([
                'Eventcorrelations.host_id IN' => $hostIds
            ])
            ->disableHydration();

        //FileDebugger::dieQuery($query);

        $hosts = $query->all();
        if ($hosts === null) {
            return [];
        }

        return $hosts->toArray();
    }

    /**
     * Get virtual services of >layer1 of an Eventcorrelation
     *
     * @param int $host_id the evc host id
     * @return array        list of services
     */
    public function getCurrentEvcConfig($host_id) {
        $query = $this->find()
            ->select([
                'Eventcorrelations.id',
                'Eventcorrelations.service_id',
            ])
            ->where([
                'AND' => [
                    'Eventcorrelations.host_id' => $host_id,
                    'Eventcorrelations.operator IS NOT NULL' //consider only virtual services
                ]
            ])
            ->disableHydration();

        $services = $query->all();

        if ($services === null) {
            return [];
        }

        $services = $services->toArray();
        return Hash::combine($services, '{n}.id', '{n}.service_id');
    }

    /**
     * @param int $ownHostId
     * @param array $serviceIds
     * @return array
     */
    public function servicesUsedByList($ownHostId, $serviceIds = []) {
        $usedByList = [];
        $serviceIdsToEvcHostNames = [];
        foreach ($serviceIds as $eventCorrelationId => $serviceId) {
            $query = $this->find()
                ->select([
                    'Eventcorrelations.id',
                    'Eventcorrelations.service_id',
                    'Hosts.name',
                    'VirtualServices.lft',
                    'VirtualServices.rght',
                ])
                ->join([
                    [
                        'table'      => 'hosts',
                        'alias'      => 'Hosts',
                        'type'       => 'INNER',
                        'conditions' => [
                            'Hosts.id = Eventcorrelations.host_id'
                        ]
                    ],
                    [
                        'table'      => 'eventcorrelations',
                        'alias'      => 'VirtualServices',
                        'type'       => 'INNER',
                        'conditions' => [
                            'AND' => [
                                'VirtualServices.host_id' => $ownHostId,
                                'VirtualServices.service_id = Eventcorrelations.service_id'
                            ]
                        ]
                    ],
                ])
                ->where([
                    'NOT'                          => [
                        'Eventcorrelations.host_id' => $ownHostId,
                    ],
                    'Eventcorrelations.service_id' => $serviceId,
                ])
                ->disableHydration();


            $result = $query->all();

            if (!empty($result)) {
                $result = $result->toArray();
                $usedByList[$serviceId] = $result;
            }
        }

        $usedByList = Hash::filter($usedByList);

        if (!empty($usedByList)) {
            $query = $this->find()
                ->select([
                    'Eventcorrelations.service_id',
                    'Eventcorrelations.lft',
                    'Eventcorrelations.rght'
                ])
                ->where([
                    'Eventcorrelations.host_id' => $ownHostId,
                    'Eventcorrelations.operator IS NOT NULL'
                ])
                ->disableHydration();

            $eventCorrelations = $query->all();
            if (empty($eventCorrelations)) {
                $eventCorrelations = [];
            } else {
                $eventCorrelations = $eventCorrelations->toArray();
            }

            foreach ($usedByList as $serviceId => $usedEvcServices) {
                foreach ($usedEvcServices as $key => $usedEvcService) {
                    if (!empty($serviceIdsToEvcHostNames[$serviceId]) &&
                        in_array($usedEvcService['Hosts']['name'], $serviceIdsToEvcHostNames[$serviceId], true)
                    ) {
                        continue;
                    }
                    $serviceIdsToEvcHostNames[$serviceId][] = $usedEvcService['Hosts']['name'];
                    foreach ($eventCorrelations as $eventCorrelation) {
                        if ($eventCorrelation['lft'] > $usedEvcService['VirtualServices']['lft'] && $eventCorrelation['lft'] < $usedEvcService['VirtualServices']['rght']) {
                            if ($eventCorrelation['lft'] > $usedEvcService['VirtualServices']['rght']) {
                                break;
                            }
                            if (!empty($serviceIdsToEvcHostNames[$eventCorrelation['service_id']]) &&
                                in_array($usedEvcService['Hosts']['name'], $serviceIdsToEvcHostNames[$eventCorrelation['service_id']])
                            ) {
                                break;
                            }
                            $serviceIdsToEvcHostNames[$eventCorrelation['service_id']][] = $usedEvcService['Hosts']['name'];
                        }
                    }
                }
            }
        }
        return $serviceIdsToEvcHostNames;
    }

    /**
     * @param int $serviceId
     * @param int $evcHostId
     * @return bool
     */
    public function isServiceUsedByOtherEvcs($serviceId, $evcHostId) {
        $count = $this->find()
            ->where([
                'Eventcorrelations.service_id' => $serviceId,
                'NOT'                          => [
                    'Eventcorrelations.host_id' => $evcHostId,
                ]
            ])
            ->count();

        return $count > 0;
    }

    /**
     * @param int $evcHostId
     */
    public function setTreeScope($evcHostId) {
        $this->behaviors()->Tree->setConfig('scope', [
            'Eventcorrelations.host_id' => $evcHostId
        ]);
    }

    /**
     * @param int|string $serviceId
     * @return bool
     */
    public function isNewEvcService($serviceId) {
        //Existing services, have an numeric service id (Auto Increment of services table)
        //New services have an fake uuid uu-id-$UUID$_vService

        return !is_numeric($serviceId);
    }

    /**
     * @param int $evcHostId
     * @return array
     */
    public function getServiceIdsByEvcHostId($evcHostId) {
        $query = $this->find('list', [
            'keyField'   => 'id',
            'valueField' => 'service_id'
        ])
            ->where([
                'Eventcorrelations.host_id' => $evcHostId,
                'Eventcorrelations.operator IS NOT NULL' //consider only virtual services
            ])
            ->disableHydration();

        return $query->toArray();
    }

    /**
     * @param int|null $evcHostId
     */
    public function cleanUpEventCorrelation($evcHostId = null) {
        if ($evcHostId === null) {
            throw new \RuntimeException('$evcHostId can not be null');
        }

        /** @var $ServicesTable ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        $firstLayerHosts = $this->getFirstLayerServicesByEvcHostId($evcHostId);
        $fistLayerServiceIds = array_unique(Hash::extract($firstLayerHosts, '{n}.service.id'));

        foreach ($fistLayerServiceIds as $fistLayerServiceId) {
            if ($this->isServiceUsedByOtherEvcs($fistLayerServiceId, $evcHostId) === false) {
                //Remove EVC Usage Flag from service
                $ServicesTable->removeUsageFlagById($fistLayerServiceId, EVENTCORRELATION_MODULE);
            }
        }

        //Delete complete EVC Tree
        $this->deleteAll([
            'Eventcorrelation.host_id' => $evcHostId
        ]);
    }

    /**
     * @param int|null $hostId
     * @return array
     */
    public function getAllServiceIdsByHostIdAndEvcUsageFlag($hostId = null) {
        if ($hostId === null) {
            throw new \RuntimeException('$hostId can not be null');
        }

        /** @var $ServicesTable ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        $query = $ServicesTable->find('list', [
            'valueField' => 'id'
        ])
            ->where([
                'Services.host_id' => $hostId,
                'Services.usage_flag &' . EVENTCORRELATION_MODULE
            ])
            ->disableHydration();

        return $query->toArray();
    }

    /**
     * @param array $serviceIds
     * @return array
     */
    public function getAllEvcsByServiceIds($serviceIds = []) {
        if (empty($serviceIds)) {
            return [];
        }
        if (!is_array($serviceIds)) {
            $serviceIds = [$serviceIds];
        }
        $query = $this->find()
            ->select([
                'Hosts.id',
                'Hosts.name',
                'Hosts.description',
                'Hosts.container_id'
            ])
            ->disableAutoFields()
            ->leftJoin(
                ['Hosts' => 'hosts'],
                ['Hosts.id = Eventcorrelations.host_id']
            )
            ->where([
                'Eventcorrelations.service_id IN' => $serviceIds
            ])
            ->group(['Hosts.id'])
            ->disableHydration();

        $eventcorrelationsAsList = $query->toArray();
        if (empty($eventcorrelationsAsList)) {
            $eventcorrelationsAsList = [];
        }
        return $eventcorrelationsAsList;
    }
}
